#
# This file is part of snmpfwd software.
#
# Copyright (c) 2014-2018, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpfwd/license.html
#
from pyasn1.type import univ, namedtype, namedval
from pyasn1.codec.ber import encoder, decoder
from pyasn1.error import SubstrateUnderrunError
from pysnmp.proto import rfc1905
from snmpfwd.trunking import crypto
from snmpfwd.error import SnmpfwdError

PROTOCOL_VERSION = 3

MSG_TYPE_REQUEST = 0
MSG_TYPE_RESPONSE = 1
MSG_TYPE_ANNOUNCEMENT = 2
MSG_TYPE_PING = 3
MSG_TYPE_PONG = 4


class Message(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('version', univ.Integer()),
        namedtype.NamedType('msg-id', univ.Integer()),
        namedtype.NamedType('content-id', univ.Integer(namedValues=namedval.NamedValues(('request', MSG_TYPE_REQUEST), ('response', MSG_TYPE_RESPONSE), ('announcement', MSG_TYPE_ANNOUNCEMENT), ('ping', MSG_TYPE_PING), ('pong', MSG_TYPE_PONG)))),
        namedtype.NamedType('payload', univ.OctetString())
    )


class Request(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('callflow-id', univ.OctetString()),
        # SNMP details
        namedtype.NamedType('snmp-engine-id', univ.OctetString()),
        namedtype.NamedType('snmp-transport-domain', univ.ObjectIdentifier()),
        namedtype.NamedType('snmp-peer-address', univ.OctetString()),
        namedtype.NamedType('snmp-peer-port', univ.Integer()),
        namedtype.NamedType('snmp-bind-address', univ.OctetString()),
        namedtype.NamedType('snmp-bind-port', univ.Integer()),
        namedtype.NamedType('snmp-security-model', univ.Integer()),
        namedtype.NamedType('snmp-security-level', univ.Integer()),
        namedtype.NamedType('snmp-security-name', univ.OctetString()),
        namedtype.NamedType('snmp-security-engine-id', univ.OctetString()),
        namedtype.NamedType('snmp-context-engine-id', univ.OctetString()),
        namedtype.NamedType('snmp-context-name', univ.OctetString()),
        namedtype.NamedType('snmp-pdu', univ.OctetString()),
        # server classifiers
        namedtype.NamedType('snmp-credentials-id', univ.OctetString('')),
        namedtype.NamedType('snmp-context-id', univ.OctetString('')),
        namedtype.NamedType('snmp-content-id', univ.OctetString('')),
        namedtype.NamedType('snmp-peer-id', univ.OctetString('')),
    )


class Response(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('error-indication', univ.OctetString('')),
        namedtype.NamedType('snmp-pdu', univ.OctetString(''))
    )


class Announcement(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('trunk-id', univ.OctetString())
    )


class Ping(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('serial', univ.Integer())
    )


class Pong(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('serial', univ.Integer())
    )


pduMap = {
    MSG_TYPE_REQUEST: Request(),
    MSG_TYPE_RESPONSE: Response(),
    MSG_TYPE_ANNOUNCEMENT: Announcement(),
    MSG_TYPE_PING: Ping(),
    MSG_TYPE_PONG: Pong(),
}


def prepareRequestData(msgId, req, secret):
    r = Request()

    r['callflow-id'] = req['callflow-id']

    for k in ('snmp-engine-id',
              'snmp-transport-domain',
              'snmp-peer-address',
              'snmp-peer-port',
              'snmp-bind-address',
              'snmp-bind-port',
              'snmp-security-model',
              'snmp-security-level',
              'snmp-security-name',
              'snmp-security-engine-id',
              'snmp-context-engine-id',
              'snmp-context-name',
              'snmp-credentials-id',
              'snmp-context-id',
              'snmp-content-id',
              'snmp-peer-id'):
        r[k] = req[k]

    r['snmp-pdu'] = encoder.encode(req['snmp-pdu'])

    msg = Message()
    msg['version'] = PROTOCOL_VERSION
    msg['msg-id'] = msgId
    msg['content-id'] = 'request'

    if secret:
        msg['payload'] = crypto.encrypt(secret, encoder.encode(r))
    else:
        msg['payload'] = encoder.encode(r)

    return encoder.encode(msg)


def prepareResponseData(msgId, rsp, secret):
    r = Response()
    r['error-indication'] = str(rsp.get('error-indication', ''))
    r['snmp-pdu'] = rsp['snmp-pdu'] and encoder.encode(rsp['snmp-pdu']) or ''

    msg = Message()
    msg['version'] = PROTOCOL_VERSION
    msg['msg-id'] = msgId
    msg['content-id'] = 'response'
    msg['payload'] = encoder.encode(r)
    if secret:
        msg['payload'] = crypto.encrypt(secret, encoder.encode(r))
    else:
        msg['payload'] = encoder.encode(r)
    return encoder.encode(msg)


def prepareAnnouncementData(trunkId, secret):
    r = Announcement()
    r['trunk-id'] = trunkId

    msg = Message()
    msg['version'] = PROTOCOL_VERSION
    msg['msg-id'] = 0
    msg['content-id'] = 'announcement'
    msg['payload'] = encoder.encode(r)
    if secret:
        msg['payload'] = crypto.encrypt(secret, encoder.encode(r))
    else:
        msg['payload'] = encoder.encode(r)
    return encoder.encode(msg)


def _preparePingPongData(reqType, msgId, serial, secret):
    msg = Message()
    msg['version'] = PROTOCOL_VERSION
    msg['msg-id'] = msgId
    msg['content-id'] = reqType

    if msg['content-id'] == MSG_TYPE_PING:
        r = Ping()
    elif msg['content-id'] == MSG_TYPE_PONG:
        r = Pong()
    else:
        raise SnmpfwdError('not a ping-pong message')

    r['serial'] = serial

    msg['payload'] = encoder.encode(r)
    if secret:
        msg['payload'] = crypto.encrypt(secret, encoder.encode(r))
    else:
        msg['payload'] = encoder.encode(r)
    return encoder.encode(msg)


def preparePingData(msgId, serial, secret):
    return _preparePingPongData(MSG_TYPE_PING, msgId, serial, secret)


def preparePongData(msgId, serial, secret):
    return _preparePingPongData(MSG_TYPE_PONG, msgId, serial, secret)


def prepareDataElements(octets, secret):
    try:
        msg, octets = decoder.decode(octets, asn1Spec=Message())

    except SubstrateUnderrunError:
        return None, None, None, octets

    if msg['version'] != PROTOCOL_VERSION:
        raise SnmpfwdError('Unsupported protocol versions %s/%s' % (msg['version'], PROTOCOL_VERSION))

    r, _ = decoder.decode(
        secret and crypto.decrypt(secret, msg['payload'].asOctets()) or msg['payload'].asOctets(),
        asn1Spec=pduMap.get(msg['content-id'])
    )

    rsp = {}

    if msg['content-id'] == MSG_TYPE_REQUEST:
        rsp['callflow-id'] = r['callflow-id']

        for k in ('snmp-engine-id',
                  'snmp-transport-domain',
                  'snmp-peer-address',
                  'snmp-peer-port',
                  'snmp-bind-address',
                  'snmp-bind-port',
                  'snmp-security-model',
                  'snmp-security-level',
                  'snmp-security-name',
                  'snmp-security-engine-id',
                  'snmp-context-engine-id',
                  'snmp-context-name',
                  'snmp-credentials-id',
                  'snmp-context-id',
                  'snmp-content-id',
                  'snmp-peer-id'):
            rsp[k] = r[k]

        pdu, _ = decoder.decode(r['snmp-pdu'], asn1Spec=rfc1905.PDUs())
        rsp['snmp-pdu'] = pdu.getComponent()

    elif msg['content-id'] == MSG_TYPE_RESPONSE:
        rsp['error-indication'] = r['error-indication']
        if not r['error-indication'] and r['snmp-pdu']:
            pdu, _ = decoder.decode(r['snmp-pdu'], asn1Spec=rfc1905.PDUs())
            rsp['snmp-pdu'] = pdu.getComponent()

    elif msg['content-id'] == MSG_TYPE_ANNOUNCEMENT:
        rsp['trunk-id'] = r['trunk-id']

    elif msg['content-id'] == MSG_TYPE_PING:
        rsp['serial'] = r['serial']

    elif msg['content-id'] == MSG_TYPE_PONG:
        rsp['serial'] = r['serial']

    return msg['msg-id'], msg['content-id'], rsp, octets



#
# This file is part of snmpfwd software.
#
# Copyright (c) 2014-2018, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpfwd/license.html
#
from snmpfwd.trunking import client, server
from snmpfwd import log, error


class TrunkingManager(object):
    def __init__(self, dataCbFun):
        self.__clients = {}
        self.__runningServersTrunkMap = {}
        self.__runningServersConnMap = {}
        self.__runningClientsTrunkMap = {}
        self.__runningClientsConnMap = {}
        self.__pingPeriods = {}
        self.__upcomingPings = {}
        self.__unconfirmedPings = {}
        self.__serial = 0
        self.__dataCbFun = dataCbFun

    def sendReq(self, trunkId, req, cbFun, cbCtx):
        if trunkId in self.__runningServersTrunkMap:
            trunk = self.__runningServersTrunkMap[trunkId]
        elif trunkId in self.__runningClientsTrunkMap:
            trunk = self.__runningClientsTrunkMap[trunkId]
        else:
            raise error.SnmpfwdError('Unknown trunk ID %s' % trunkId)

        return trunk.sendReq(req, cbFun, cbCtx)

    def sendRsp(self, trunkId, msgId, rsp):
        if trunkId in self.__runningServersTrunkMap:
            trunk = self.__runningServersTrunkMap[trunkId]
        elif trunkId in self.__runningClientsTrunkMap:
            trunk = self.__runningClientsTrunkMap[trunkId]
        else:
            raise error.SnmpfwdError('Unknown trunk ID %s' % trunkId)

        return trunk.sendRsp(msgId, rsp)

    def setupTrunks(self, timeNow):
        for trunkId in self.__clients:
            if trunkId in self.__runningClientsTrunkMap and \
                    not self.__runningClientsTrunkMap[trunkId].isUp:
                self.__runningClientsTrunkMap[trunkId].close()
                del self.__runningClientsConnMap[self.__runningClientsTrunkMap[trunkId]]
                del self.__runningClientsTrunkMap[trunkId]

            if trunkId not in self.__runningClientsTrunkMap:
                self.__runningClientsTrunkMap[trunkId] = client.TrunkingClient(
                    *self.__clients[trunkId]
                )
                self.__runningClientsConnMap[self.__runningClientsTrunkMap[trunkId]] = trunkId
                self.__runningClientsTrunkMap[trunkId].sendAnnouncement(trunkId)

    def __monitorTrunksCbFun(self, msg, cbCtx):
        trunkId, expectedSerial = cbCtx
        if msg['serial'] == expectedSerial:
            if trunkId in self.__unconfirmedPings:
                del self.__unconfirmedPings[trunkId]

    def __monitorTrunks(self, timeNow, runningConnections, runningConnectionsMap):
        for trunkId in tuple(runningConnections):
            pingPeriod = self.__pingPeriods[trunkId]
            if not pingPeriod:
                continue

            nextPingAt = self.__upcomingPings.get(trunkId, timeNow)

            if nextPingAt > timeNow:
                continue

            self.__upcomingPings[trunkId] = timeNow + pingPeriod

            connection = runningConnections[trunkId]

            if trunkId in self.__unconfirmedPings:
                log.error('closing unresponsive trunk %s %s' % (trunkId, connection))
                connection.close()
                del runningConnectionsMap[connection]
                del runningConnections[trunkId]
                del self.__unconfirmedPings[trunkId]
                continue

            connection.sendPing(self.__serial, self.__monitorTrunksCbFun, (trunkId, self.__serial))
            self.__unconfirmedPings[trunkId] = True

        self.__serial += 1

    def monitorTrunks(self, timeNow):
        self.__monitorTrunks(
            timeNow, self.__runningClientsTrunkMap, self.__runningClientsConnMap
        )
        self.__monitorTrunks(
            timeNow, self.__runningServersTrunkMap, self.__runningServersConnMap
        )

    def addClient(self, trunkId, localEndpoint, remoteEndpoint, pingPeriod, secret):
        if trunkId in self.__clients or trunkId in self.__runningServersTrunkMap:
            raise error.SnmpfwdError('Trunk %s already registered' % trunkId)
        self.__clients[trunkId] = localEndpoint, remoteEndpoint, secret, self.__proxyDataCbFun
        self.__pingPeriods[trunkId] = pingPeriod

    def __proxyDataCbFun(self, connection, msgId, msg):
        if connection in self.__runningServersConnMap:
            trunkId = self.__runningServersConnMap[connection]
        elif connection in self.__runningClientsConnMap:
            trunkId = self.__runningClientsConnMap[connection]
        else:
            log.error('data message from unknown connection %s ignored' % connection)
            return

        self.__dataCbFun(trunkId, msgId, msg)
        
    def __ctlCbFun(self, connection, msg, cbCtx):
        if msg:
            trunkId = str(msg['trunk-id'])
            pingPeriod = cbCtx
            if trunkId in self.__runningServersTrunkMap or \
                    trunkId in self.__runningClientsTrunkMap:
                log.error('duplicate trunk %s during negotiation with %s' % (trunkId, connection))
                connection.close()
                return

            log.info('registering connection %s as trunk %s' % (connection, trunkId))
            self.__runningServersTrunkMap[trunkId] = connection
            self.__runningServersConnMap[connection] = trunkId
            self.__pingPeriods[trunkId] = pingPeriod

        else:
            if connection in self.__runningServersConnMap:
                trunkId = self.__runningServersConnMap[connection]
            else:
                log.error('control message from unknown connection %s ignored' % connection)
                return
                
            log.info('unregistering connection %s (trunk %s)' % (self.__runningServersTrunkMap[trunkId], trunkId))
            del self.__runningServersTrunkMap[trunkId]
            del self.__runningServersConnMap[connection]

    def addServer(self, localEndpoint, pingPeriod, secret):
        server.TrunkingSuperServer(
            localEndpoint, secret, self.__proxyDataCbFun, self.__ctlCbFun, pingPeriod
        )



#
# This file is part of snmpfwd software.
#
# Copyright (c) 2014-2018, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpfwd/license.html
#
# SNMP Proxy Forwarder plugin module
#
import logging
from logging import handlers
import time
import os
import stat
try:
    from ConfigParser import RawConfigParser, Error
except ImportError:
    from configparser import RawConfigParser, Error
from snmpfwd.plugins import status
from snmpfwd.error import SnmpfwdError
from snmpfwd.log import debug, info, error
from pysnmp.proto.api import v2c

hostProgs = 'snmpfwd-server', 'snmpfwd-client'

apiVersions = 0, 2

PLUGIN_NAME = 'logger'

# defaults
pduMap = {}
method = 'null'
logFormat = ''
leftParen, rightParen = '', ''

logger = logging.getLogger('snmpfwd-logger')

for moduleOption in moduleOptions:

    optionName, optionValue = moduleOption.split('=', 1)

    if optionName == 'config':

        configFile = optionValue

        config = RawConfigParser()
        config.read(configFile)

        method = config.get('general', 'method')
        if method == 'file':

            rotation = config.get('file', 'rotation')

            if rotation == 'timed':
                filename = config.get('file', 'destination')

                handler = handlers.TimedRotatingFileHandler(
                    filename,
                    config.get('file', 'timescale'),
                    int(config.get('file', 'interval')),
                    int(config.get('file', 'backupcount'))
                )

                # Use file metadata change time instead of last modification
                # time as the stock implementation does.
                # This is to work-around the increasing rotation intervals
                # on process restart.
                # Warning: this is based on undocumented logger features.
                if os.path.exists(filename):
                    createdAt = os.stat(filename)[stat.ST_CTIME]
                    handler.rolloverAt = handler.computeRollover(createdAt)

            else:
                raise SnmpfwdError('%s: unknown rotation method %s at %s' % (PLUGIN_NAME, rotation, configFile))

        else:
            raise SnmpfwdError('%s: unknown logging method %s at %s' % (PLUGIN_NAME, method, configFile))

        debug('%s: using %s logging method' % (PLUGIN_NAME, method))

        logger.setLevel(logging.INFO)

        pdus = config.get('content', 'pdus')
        for pdu in pdus.split():
            try:
                pduMap[getattr(v2c, pdu + 'PDU').tagSet] = pdu

            except AttributeError:
                raise SnmpfwdError('%s: unknown PDU %s' % (PLUGIN_NAME, pdu))

            else:
                debug('%s: PDU ACL includes %s' % (PLUGIN_NAME, pdu))

        try:
            leftParen, rightParen = config.get('content', 'parentheses').split()

        except Exception:
            error('%s: malformed "parentheses" values at %s' % (PLUGIN_NAME, configFile))

        template = config.get('content', 'template')

        debug('%s: using var-bind value parentheses %s %s' % (PLUGIN_NAME, leftParen, rightParen))

        logger.addHandler(handler)

started = time.time()

info('%s: plugin initialization complete' % PLUGIN_NAME)


def _format(template, pdu, context):
    for key, value in context.items():
        token = '${%s}' % key
        if token in template:
            template = template.replace(token, str(value))

    token = '${snmp-var-binds}'
    if token in template:
        varBinds = ['%s %s%s%s' % (vb[0].prettyPrint(),
                                   leftParen,
                                   vb[1].prettyPrint().replace(leftParen, leftParen + leftParen).replace(rightParen, rightParen + rightParen),
                                   rightParen)
                    for vb in v2c.apiPDU.getVarBinds(pdu)]
        template = template.replace(token, ' '.join(varBinds))

    token = '${snmp-pdu-type}'
    if token in template:
        template = template.replace(token, pduMap[pdu.tagSet])

    now = time.time()

    token = '${asctime}'
    if token in template:
        timestamp = time.asctime(time.localtime(now))
        template = template.replace(token, timestamp)

    token = '${isotime}'
    if token in template:
        timestamp = time.strftime('%Y-%m-%dT%H:%M:%S', time.localtime(now))
        timestamp += '.%02d' % (now % 1 * 100)
        template = template.replace(token, timestamp)

    token = '${timestamp}'
    if token in template:
        timestamp = '%.2f' % now
        template = template.replace(token, timestamp)

    token = '${uptime}'
    if token in template:
        timestamp = '%012.2f' % (time.time() - started)
        template = template.replace(token, timestamp)

    return template


def processCommandRequest(pluginId, snmpEngine, pdu, trunkMsg, reqCtx):
    if pdu.tagSet in pduMap:
        logger.info(_format(template, pdu, trunkMsg))

    return status.NEXT, pdu

processCommandResponse = processCommandRequest

processNotificationRequest = processCommandRequest

processNotificationResponse = processCommandRequest



#!/usr/bin/env python
"""SNMP Proxy Forwarder can act as an application-level firewall
   or SNMP protocol translator that let SNMPv1/v2c entities to talk
   to SNMPv3 ones or vice-versa.
"""
import sys
import os
import glob

classifiers = """\
Development Status :: 5 - Production/Stable
Environment :: Console
Intended Audience :: Developers
Intended Audience :: Education
Intended Audience :: Information Technology
Intended Audience :: System Administrators
Intended Audience :: Telecommunications Industry
License :: OSI Approved :: BSD License
Natural Language :: English
Operating System :: OS Independent
Programming Language :: Python :: 2
Programming Language :: Python :: 2.4
Programming Language :: Python :: 2.5
Programming Language :: Python :: 2.6
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.2
Programming Language :: Python :: 3.3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Topic :: Communications
Topic :: System :: Monitoring
Topic :: System :: Networking :: Monitoring
"""

def howto_install_setuptools():
    print("""
   Error: You need setuptools Python package!

   It's very easy to install it, just type:

   wget https://bootstrap.pypa.io/ez_setup.py
   python ez_setup.py

   Then you could make eggs from this package.
""")

if sys.version_info[:2] < (2, 4):
    print("ERROR: this package requires Python 2.4 or later!")
    sys.exit(1)

try:
    from setuptools import setup

    params = {
        'install_requires': ['pysnmp>=4.4.3', 'pycryptodomex'],
        'zip_safe': True
    }

except ImportError:
    for arg in sys.argv:
        if 'egg' in arg:
            howto_install_setuptools()
            sys.exit(1)

    from distutils.core import setup

    params = {}
    if sys.version_info[:2] > (2, 4):
        params['requires'] = ['pysnmp(>=4.4.3)', 'pycryptodomex']

doclines = [x.strip() for x in (__doc__ or '').split('\n') if x]

params.update(
    {'name': "snmpfwd",
     'version':  open(os.path.join('snmpfwd', '__init__.py')).read().split('\'')[1],
     'description': doclines[0],
     'long_description': ' '.join(doclines[1:]),
     'maintainer': 'Ilya Etingof <etingof@gmail.com>',
     'author': "Ilya Etingof",
     'author_email': "etingof@gmail.com",
     'url': "https://github.com/etingof/snmpfwd",
     'platforms': ['any'],
     'classifiers': [x for x in classifiers.split('\n') if x],
     'scripts': ['scripts/snmpfwd-client.py', 'scripts/snmpfwd-server.py'],
     'packages': ['snmpfwd', 'snmpfwd.trunking', 'snmpfwd.plugins'],
     'license': "BSD"}
)


params['data_files'] = [
    ( 'snmpfwd/' + 'plugins', glob.glob(os.path.join('plugins', '*.py')) )
]

if 'py2exe' in sys.argv:
    import py2exe

    # fix executables
    params['console'] = params['scripts']
    del params['scripts']

    # pysnmp used by snmpfwd dynamically loads some of its *.py files
    params['options'] = {
        'py2exe': {
            'includes': [
                'pysnmp.smi.mibs.*',
                'pysnmp.smi.mibs.instances.*'
            ],
            'bundle_files': 1,
            'compressed': True
        }
    }

    params['zipfile'] = None

    del params['data_files']  # no need to store these in .exe

    # additional modules used by snmpfwd but not seen by py2exe
    for m in ('random',):
        try:
            __import__(m)
        except ImportError:
            continue
        else:
            params['options']['py2exe']['includes'].append(m)

    print("!!! Make sure your pysnmp/pyasn1 packages are NOT .egg'ed!!!")

setup(**params)
